@@ -4,10 +4,11 @@ import (
 	"context"
 	"errors"
 	"fmt"
+	"reflect"
 	"sync"
 	"time"
 
-	"k8s.io/klog"
+	"k8s.io/klog/v2"
 
 	corev1 "k8s.io/api/core/v1"
 	networkingv1beta1 "k8s.io/api/networking/v1beta1"
@@ -25,6 +26,7 @@ import (
 	coreinformers "k8s.io/client-go/informers/core/v1"
 	networkingv1beta1informers "k8s.io/client-go/informers/networking/v1beta1"
 	kv1core "k8s.io/client-go/kubernetes/typed/core/v1"
+	kv1beta1networking "k8s.io/client-go/kubernetes/typed/networking/v1beta1"
 	corelisters "k8s.io/client-go/listers/core/v1"
 	networkingv1beta1listers "k8s.io/client-go/listers/networking/v1beta1"
 	"k8s.io/client-go/tools/cache"
@@ -73,7 +75,8 @@ import (
 type Controller struct {
 	eventRecorder record.EventRecorder
 
-	client routeclient.RoutesGetter
+	routeClient   routeclient.RoutesGetter
+	ingressClient kv1beta1networking.IngressesGetter
 
 	ingressLister networkingv1beta1listers.IngressLister
 	secretLister  corelisters.SecretLister
@@ -160,7 +163,7 @@ type queueKey struct {
 }
 
 // NewController instantiates a Controller
-func NewController(eventsClient kv1core.EventsGetter, client routeclient.RoutesGetter, ingresses networkingv1beta1informers.IngressInformer, secrets coreinformers.SecretInformer, services coreinformers.ServiceInformer, routes routeinformers.RouteInformer) *Controller {
+func NewController(eventsClient kv1core.EventsGetter, routeClient routeclient.RoutesGetter, ingressClient kv1beta1networking.IngressesGetter, ingresses networkingv1beta1informers.IngressInformer, secrets coreinformers.SecretInformer, services coreinformers.ServiceInformer, routes routeinformers.RouteInformer) *Controller {
 	broadcaster := record.NewBroadcaster()
 	broadcaster.StartLogging(klog.Infof)
 	// TODO: remove the wrapper when every clients have moved to use the clientset.
@@ -174,7 +177,8 @@ func NewController(eventsClient kv1core.EventsGetter, client routeclient.RoutesG
 		expectations:     newExpectations(),
 		expectationDelay: 2 * time.Second,
 
-		client: client,
+		routeClient:   routeClient,
+		ingressClient: ingressClient,
 
 		ingressLister: ingresses.Lister(),
 		secretLister:  secrets.Lister(),
@@ -382,7 +386,7 @@ func (c *Controller) sync(key queueKey) error {
 
 	// walk the ingress and identify whether any of the child routes need to be updated, deleted,
 	// or created, as efficiently as possible.
-	var creates, updates []*routev1.Route
+	var creates, updates, matches []*routev1.Route
 	for _, rule := range ingress.Spec.Rules {
 		if rule.HTTP == nil {
 			continue
@@ -405,6 +409,7 @@ func (c *Controller) sync(key queueKey) error {
 			}
 
 			if routeMatchesIngress(existing, ingress, &rule, &path, c.secretLister, c.serviceLister) {
+				matches = append(matches, existing)
 				continue
 			}
 
@@ -423,7 +428,7 @@ func (c *Controller) sync(key queueKey) error {
 
 	// add the new routes
 	for _, route := range creates {
-		if err := createRouteWithName(c.client, ingress, route, c.expectations); err != nil {
+		if err := createRouteWithName(c.routeClient, ingress, route, c.expectations); err != nil {
 			errs = append(errs, err)
 		}
 	}
@@ -435,7 +440,7 @@ func (c *Controller) sync(key queueKey) error {
 			return err
 		}
 		data = []byte(fmt.Sprintf(`[{"op":"replace","path":"/spec","value":%s}]`, data))
-		_, err = c.client.Routes(route.Namespace).Patch(context.TODO(), route.Name, types.JSONPatchType, data, metav1.PatchOptions{})
+		_, err = c.routeClient.Routes(route.Namespace).Patch(context.TODO(), route.Name, types.JSONPatchType, data, metav1.PatchOptions{})
 		if err != nil {
 			errs = append(errs, err)
 		}
@@ -443,15 +448,63 @@ func (c *Controller) sync(key queueKey) error {
 
 	// purge any previously managed routes
 	for _, route := range old {
-		if err := c.client.Routes(route.Namespace).Delete(context.TODO(), route.Name, metav1.DeleteOptions{}); err != nil && !kerrors.IsNotFound(err) {
+		if err := c.routeClient.Routes(route.Namespace).Delete(context.TODO(), route.Name, metav1.DeleteOptions{}); err != nil && !kerrors.IsNotFound(err) {
 			errs = append(errs, err)
 		}
 	}
 
-	if len(errs) > 0 {
-		return utilerrors.NewAggregate(errs)
+	// reflect route status to ingress status
+	//
+	// We must preserve status that other controllers may have added, so we
+	// cannot simply compute the new status from the routes associated with
+	// the ingress; instead, we need to take the current status, remove
+	// hosts for routes we've just deleted, and then add hosts for current
+	// routes.  In sum, we compute
+	// ingress.Status.LoadBalancer.Ingress[*].Hostname -
+	// old[*].Status.Ingress[*].RouterCanonicalHostname +
+	// matches[*].Status.Ingress[*].RouterCanonicalHostname.
+	oldCanonicalHostnames := sets.NewString()
+	for _, ingressIngress := range ingress.Status.LoadBalancer.Ingress {
+		oldCanonicalHostnames.Insert(ingressIngress.Hostname)
 	}
-	return nil
+	newCanonicalHostnames := sets.NewString(oldCanonicalHostnames.List()...)
+	for _, route := range old {
+		for _, routeIngress := range route.Status.Ingress {
+			for _, cond := range routeIngress.Conditions {
+				if cond.Type == routev1.RouteAdmitted {
+					if cond.Status == corev1.ConditionTrue {
+						newCanonicalHostnames.Delete(routeIngress.RouterCanonicalHostname)
+					}
+					break
+				}
+			}
+		}
+	}
+	for _, route := range matches {
+		for _, routeIngress := range route.Status.Ingress {
+			for _, cond := range routeIngress.Conditions {
+				if cond.Type == routev1.RouteAdmitted {
+					if cond.Status == corev1.ConditionTrue {
+						newCanonicalHostnames.Insert(routeIngress.RouterCanonicalHostname)
+					}
+					break
+				}
+			}
+		}
+	}
+	if !newCanonicalHostnames.Equal(oldCanonicalHostnames) {
+		updatedIngress := ingress.DeepCopy()
+		ingressIngresses := make([]corev1.LoadBalancerIngress, 0, newCanonicalHostnames.Len())
+		for _, canonicalHostname := range newCanonicalHostnames.List() {
+			ingressIngresses = append(ingressIngresses, corev1.LoadBalancerIngress{Hostname: canonicalHostname})
+		}
+		updatedIngress.Status.LoadBalancer.Ingress = ingressIngresses
+		if _, err := c.ingressClient.Ingresses(key.namespace).UpdateStatus(context.TODO(), updatedIngress, metav1.UpdateOptions{}); err != nil {
+			errs = append(errs, err)
+		}
+	}
+
+	return utilerrors.NewAggregate(errs)
 }
 
 func hasIngressOwnerRef(owners []metav1.OwnerReference) (string, bool) {
@@ -471,37 +524,17 @@ func newRouteForIngress(
 	secretLister corelisters.SecretLister,
 	serviceLister corelisters.ServiceLister,
 ) *routev1.Route {
-	var tlsConfig *routev1.TLSConfig
-	if name, ok := referencesSecret(ingress, rule.Host); ok {
-		secret, err := secretLister.Secrets(ingress.Namespace).Get(name)
-		if err != nil {
-			// secret doesn't exist yet, wait
-			return nil
-		}
-		if secret.Type != corev1.SecretTypeTLS {
-			// secret is the wrong type
-			return nil
-		}
-		if _, ok := secret.Data[corev1.TLSCertKey]; !ok {
-			return nil
-		}
-		if _, ok := secret.Data[corev1.TLSPrivateKeyKey]; !ok {
-			return nil
-		}
-		tlsConfig = &routev1.TLSConfig{
-			Termination:                   routev1.TLSTerminationEdge,
-			Certificate:                   string(secret.Data[corev1.TLSCertKey]),
-			Key:                           string(secret.Data[corev1.TLSPrivateKeyKey]),
-			InsecureEdgeTerminationPolicy: routev1.InsecureEdgeTerminationPolicyRedirect,
-		}
-	}
-
 	targetPort, err := targetPortForService(ingress.Namespace, path, serviceLister)
 	if err != nil {
 		// no valid target port
 		return nil
 	}
 
+	tlsSecret, hasInvalidTLSSecret := tlsSecretIfValid(ingress, rule, secretLister)
+	if hasInvalidTLSSecret {
+		return nil
+	}
+
 	var port *routev1.RoutePort
 	if targetPort != nil {
 		port = &routev1.RoutePort{TargetPort: *targetPort}
@@ -525,7 +558,7 @@ func newRouteForIngress(
 				Name: path.Backend.ServiceName,
 			},
 			Port: port,
-			TLS:  tlsConfig,
+			TLS:  tlsConfigForIngress(ingress, rule, tlsSecret),
 		},
 	}
 }
@@ -570,17 +603,15 @@ func routeMatchesIngress(
 		return false
 	}
 
-	var secret *corev1.Secret
-	if name, ok := referencesSecret(ingress, rule.Host); ok {
-		secret, _ = secretLister.Secrets(ingress.Namespace).Get(name)
-		if secret == nil {
-			return false
-		}
-	}
-	if !secretMatchesRoute(secret, route.Spec.TLS) {
+	tlsSecret, hasInvalidTLSSecret := tlsSecretIfValid(ingress, rule, secretLister)
+	if hasInvalidTLSSecret {
 		return false
 	}
-	return true
+	tlsConfig := tlsConfigForIngress(ingress, rule, tlsSecret)
+	if route.Spec.TLS != nil && tlsConfig != nil {
+		tlsConfig.InsecureEdgeTerminationPolicy = route.Spec.TLS.InsecureEdgeTerminationPolicy
+	}
+	return reflect.DeepEqual(tlsConfig, route.Spec.TLS)
 }
 
 // targetPortForService returns a target port for a Route based on the given
@@ -625,27 +656,6 @@ func targetPortForService(namespace string, path *networkingv1beta1.HTTPIngressP
 	return nil, errors.New("no port found")
 }
 
-func secretMatchesRoute(secret *corev1.Secret, tlsConfig *routev1.TLSConfig) bool {
-	if secret == nil {
-		return tlsConfig == nil
-	}
-	if secret.Type != corev1.SecretTypeTLS {
-		return tlsConfig == nil
-	}
-	if _, ok := secret.Data[corev1.TLSCertKey]; !ok {
-		return false
-	}
-	if _, ok := secret.Data[corev1.TLSPrivateKeyKey]; !ok {
-		return false
-	}
-	if tlsConfig == nil {
-		return false
-	}
-	return tlsConfig.Termination == routev1.TLSTerminationEdge &&
-		tlsConfig.Certificate == string(secret.Data[corev1.TLSCertKey]) &&
-		tlsConfig.Key == string(secret.Data[corev1.TLSPrivateKeyKey])
-}
-
 func splitForPathAndHost(routes []*routev1.Route, host, path string) ([]*routev1.Route, *routev1.Route) {
 	for i, route := range routes {
 		if route.Spec.Host == host && route.Spec.Path == path {
@@ -719,3 +729,80 @@ func generateRouteName(base string) string {
 	}
 	return fmt.Sprintf("%s%s", base, utilrand.String(randomLength))
 }
+
+func tlsConfigForIngress(
+	ingress *networkingv1beta1.Ingress,
+	rule *networkingv1beta1.IngressRule,
+	potentiallyNilTLSSecret *corev1.Secret,
+) *routev1.TLSConfig {
+	if !tlsEnabled(ingress, rule, potentiallyNilTLSSecret) {
+		return nil
+	}
+	// Edge: May have cert
+	// Re-Encrypt: May have cert
+	// Passthrough: Must not have cert
+	terminationPolicy := terminationPolicyForIngress(ingress)
+	tlsConfig := &routev1.TLSConfig{
+		Termination:                   terminationPolicy,
+		InsecureEdgeTerminationPolicy: routev1.InsecureEdgeTerminationPolicyRedirect,
+	}
+	if terminationPolicy != routev1.TLSTerminationPassthrough && potentiallyNilTLSSecret != nil {
+		tlsConfig.Certificate = string(potentiallyNilTLSSecret.Data[corev1.TLSCertKey])
+		tlsConfig.Key = string(potentiallyNilTLSSecret.Data[corev1.TLSPrivateKeyKey])
+	}
+	return tlsConfig
+}
+
+var emptyTLS = networkingv1beta1.IngressTLS{}
+
+func tlsEnabled(ingress *networkingv1beta1.Ingress, rule *networkingv1beta1.IngressRule, potentiallyNilTLSSecret *corev1.Secret) bool {
+	switch ingress.Annotations[terminationPolicyAnnotationKey] {
+	case string(routev1.TLSTerminationPassthrough), string(routev1.TLSTerminationReencrypt), string(routev1.TLSTerminationEdge):
+		return true
+	}
+	if potentiallyNilTLSSecret != nil {
+		return true
+	}
+	for _, tls := range ingress.Spec.TLS {
+		if reflect.DeepEqual(tls, emptyTLS) {
+			return true
+		}
+	}
+	return false
+}
+
+func tlsSecretIfValid(ingress *networkingv1beta1.Ingress, rule *networkingv1beta1.IngressRule, secretLister corelisters.SecretLister) (_ *corev1.Secret, hasInvalidSecret bool) {
+	name, ok := referencesSecret(ingress, rule.Host)
+	if !ok {
+		return nil, false
+	}
+	secret, err := secretLister.Secrets(ingress.Namespace).Get(name)
+	if err != nil {
+		// secret doesn't exist yet, wait
+		return nil, true
+	}
+	if secret.Type != corev1.SecretTypeTLS {
+		// secret is the wrong type
+		return nil, true
+	}
+	if _, ok := secret.Data[corev1.TLSCertKey]; !ok {
+		return nil, true
+	}
+	if _, ok := secret.Data[corev1.TLSPrivateKeyKey]; !ok {
+		return nil, true
+	}
+	return secret, false
+}
+
+var terminationPolicyAnnotationKey = routev1.GroupName + "/termination"
+
+func terminationPolicyForIngress(ingress *networkingv1beta1.Ingress) routev1.TLSTerminationType {
+	switch {
+	case ingress.Annotations[terminationPolicyAnnotationKey] == string(routev1.TLSTerminationPassthrough):
+		return routev1.TLSTerminationPassthrough
+	case ingress.Annotations[terminationPolicyAnnotationKey] == string(routev1.TLSTerminationReencrypt):
+		return routev1.TLSTerminationReencrypt
+	default:
+		return routev1.TLSTerminationEdge
+	}
+}
